@@ -109,17 +109,65 @@ static void put_cpu_fault_service_context(uvm_service_block_context_t *service_c
 static int uvm_open(struct inode *inode, struct file *filp)
 {
     NV_STATUS status = uvm_global_get_status();
-    if (status == NV_OK)
+
+    if (status == NV_OK) {
+        if (!uvm_down_read_trylock(&g_uvm_global.pm.lock))
+            return -EAGAIN;
+
         status = uvm_va_space_create(inode, filp);
 
+        uvm_up_read(&g_uvm_global.pm.lock);
+    }
+
     return -nv_status_to_errno(status);
 }
 
+static void uvm_release_deferred(void *data)
+{
+    uvm_va_space_t *va_space = data;
+
+    // Since this function is only scheduled to run when uvm_release() fails
+    // to trylock-acquire the pm.lock, the following acquisition attempt
+    // is expected to block this thread, and cause it to remain blocked until
+    // uvm_resume() releases the lock.  As a result, the deferred release
+    // kthread queue may stall for long periods of time.
+    uvm_down_read(&g_uvm_global.pm.lock);
+
+    uvm_va_space_destroy(va_space);
+
+    uvm_up_read(&g_uvm_global.pm.lock);
+}
+
 static int uvm_release(struct inode *inode, struct file *filp)
 {
-    uvm_va_space_destroy(filp);
+    uvm_va_space_t *va_space = uvm_va_space_get(filp);
+    int ret;
 
-    return -nv_status_to_errno(uvm_global_get_status());
+    filp->private_data = NULL;
+    filp->f_mapping = NULL;
+
+    // Because the kernel discards the status code returned from this release
+    // callback, early exit in case of a pm.lock acquisition failure is not
+    // an option.  Instead, the teardown work normally performed synchronously
+    // needs to be scheduled to run after uvm_resume() releases the lock.
+    if (uvm_down_read_trylock(&g_uvm_global.pm.lock)) {
+        uvm_va_space_destroy(va_space);
+        uvm_up_read(&g_uvm_global.pm.lock);
+    }
+    else {
+        // Remove references to this inode from the address_space.  This isn't
+        // strictly necessary, as any CPU mappings of this file have already
+        // been destroyed, and va_space->mapping won't be used again. Still,
+        // the va_space survives the inode if its destruction is deferred, in
+        // which case the references are rendered stale.
+        address_space_init_once(&va_space->mapping);
+
+        nv_kthread_q_item_init(&va_space->deferred_release_q_item, uvm_release_deferred, va_space);
+        ret = nv_kthread_q_schedule_q_item(&g_uvm_global.deferred_release_q, &va_space->deferred_release_q_item);
+        UVM_ASSERT(ret != 0);
+    }
+
+    return 0;
 }
 
 static void uvm_destroy_vma_managed(struct vm_area_struct *vma, bool is_uvm_teardown)
@@ -404,10 +452,15 @@ static vm_fault_t uvm_vm_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
     if (status != NV_OK)
         goto convert_error;
 
+    if (!uvm_down_read_trylock(&g_uvm_global.pm.lock)) {
+        status = NV_ERR_BUSY_RETRY;
+        goto convert_error;
+    }
+
     service_context = get_cpu_fault_service_context();
     if (!service_context) {
         status = NV_ERR_NO_MEMORY;
-        goto convert_error;
+        goto unlock;
     }
 
     service_context->cpu_fault.wakeup_time_stamp = 0;
@@ -496,9 +549,13 @@ out:
     major_fault = service_context->cpu_fault.did_migrate;
     put_cpu_fault_service_context(service_context);
 
+unlock:
+    uvm_up_read(&g_uvm_global.pm.lock);
+
 convert_error:
     switch (status) {
         case NV_OK:
+        case NV_ERR_BUSY_RETRY:
             return VM_FAULT_NOPAGE | (major_fault ? VM_FAULT_MAJOR : 0);
         case NV_ERR_NO_MEMORY:
             return VM_FAULT_OOM;
@@ -623,15 +680,12 @@ static int uvm_mmap(struct file *filp, struct vm_area_struct *vma)
     if (status != NV_OK)
         return -nv_status_to_errno(status);
 
-    uvm_record_lock_mmap_sem_write(&current->mm->mmap_sem);
-
     // UVM mappings are required to set offset == VA. This simplifies things
     // since we don't have to worry about address aliasing (except for fork,
     // handled separately) and it makes unmap_mapping_range simpler.
     if (vma->vm_start != (vma->vm_pgoff << PAGE_SHIFT)) {
         UVM_DBG_PRINT_RL("vm_start 0x%lx != vm_pgoff 0x%lx\n", vma->vm_start, vma->vm_pgoff << PAGE_SHIFT);
-        ret = -EINVAL;
-        goto out;
+        return -EINVAL;
     }
 
     // Enforce shared read/writable mappings so we get all fault callbacks
@@ -640,10 +694,21 @@ static int uvm_mmap(struct file *filp, struct vm_area_struct *vma)
     if ((vma->vm_flags & (VM_SHARED|VM_READ|VM_WRITE)) !=
                          (VM_SHARED|VM_READ|VM_WRITE)) {
         UVM_DBG_PRINT_RL("User requested non-shared or non-writable mapping\n");
-        ret = -EINVAL;
-        goto out;
+        return -EINVAL;
     }
 
+    // If the PM lock cannot be acquired, disable the VMA and report success
+    // to the caller.  The caller is expected to determine whether the
+    // map operation succeeded via an ioctl() call.  This is necessary to
+    // safely handle MAP_FIXED, which needs to complete atomically to prevent
+    // the loss of the virtual address range.
+    if (!uvm_down_read_trylock(&g_uvm_global.pm.lock)) {
+        uvm_disable_vma(vma);
+        return 0;
+    }
+
+    uvm_record_lock_mmap_sem_write(&current->mm->mmap_sem);
+
     // VM_MIXEDMAP      Required to use vm_insert_page
     //
     // VM_DONTEXPAND    mremap can grow a vma in place without giving us any
@@ -711,10 +776,12 @@ out:
 
     uvm_record_unlock_mmap_sem_write(&current->mm->mmap_sem);
 
+    uvm_up_read(&g_uvm_global.pm.lock);
+
     return ret;
 }
 
-static long uvm_unlocked_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
+static long uvm_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
 {
     switch (cmd)
     {
@@ -759,12 +826,29 @@ static long uvm_unlocked_ioctl(struct file *filp, unsigned int cmd, unsigned lon
         UVM_ROUTE_CMD_ALLOC(UVM_ALLOC_SEMAPHORE_POOL,           uvm_api_alloc_semaphore_pool);
         UVM_ROUTE_CMD_STACK(UVM_CLEAN_UP_ZOMBIE_RESOURCES,      uvm_api_clean_up_zombie_resources);
         UVM_ROUTE_CMD_STACK(UVM_POPULATE_PAGEABLE,              uvm_api_populate_pageable);
+        UVM_ROUTE_CMD_STACK(UVM_VALIDATE_VA_RANGE,              uvm_api_validate_va_range);
     }
 
     // Try the test ioctls if none of the above matched
     return uvm8_test_ioctl(filp, cmd, arg);
 }
 
+static long uvm_unlocked_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
+{
+    long ret;
+
+    if (!uvm_down_read_trylock(&g_uvm_global.pm.lock))
+        return -EAGAIN;
+
+    ret = uvm_ioctl(filp, cmd, arg);
+
+    uvm_up_read(&g_uvm_global.pm.lock);
+
+    uvm_thread_assert_all_unlocked();
+
+    return ret;
+}
+
 static const struct file_operations uvm_fops =
 {
     .open            = uvm_open,
