@@ -1,5 +1,5 @@
 /*******************************************************************************
-    Copyright (c) 2015 NVIDIA Corporation
+    Copyright (c) 2015-2019 NVIDIA Corporation
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"), to
@@ -236,7 +236,7 @@ NV_STATUS uvm_channel_reserve_type(uvm_channel_manager_t *channel_manager,
 {
     unsigned ce_index;
 
-    if (type == UVM_CHANNEL_TYPE_ANY) 
+    if (type == UVM_CHANNEL_TYPE_ANY)
         return channel_reserve_in_range(channel_manager, 0, channel_manager->num_channels, channel_out);
 
     ce_index = channel_manager->ce_to_use.default_for_type[type];
@@ -512,7 +512,6 @@ static unsigned channel_pool_get_ce_index(const uvm_channel_pool_t *pool)
 static NV_STATUS create_channel(uvm_channel_pool_t *pool, bool with_procfs, uvm_channel_t *channel)
 {
     NV_STATUS status;
-    uvmGpuCopyEngineHandle ce_handle;
     uvm_channel_manager_t *manager = pool->manager;
     uvm_gpu_t *gpu = manager->gpu;
     UvmGpuChannelAllocParams channel_alloc_params;
@@ -537,10 +536,11 @@ static NV_STATUS create_channel(uvm_channel_pool_t *pool, bool with_procfs, uvm_
     channel_alloc_params.numGpFifoEntries = manager->conf.num_gpfifo_entries;
     channel_alloc_params.gpFifoLoc        = manager->conf.gpfifo_loc;
     channel_alloc_params.gpPutLoc         = manager->conf.gpput_loc;
+    channel_alloc_params.engineIndex      = ce_index;
 
     status = uvm_rm_locked_call(nvUvmInterfaceChannelAllocate(gpu->rm_address_space,
-                                                              &channel->handle,
                                                               &channel_alloc_params,
+                                                              &channel->handle,
                                                               &channel->channel_info));
     if (status != NV_OK) {
         UVM_ERR_PRINT("nvUvmInterfaceChannelAllocate() failed: %s, GPU %s\n", nvstatusToString(status), gpu->name);
@@ -553,16 +553,6 @@ static NV_STATUS create_channel(uvm_channel_pool_t *pool, bool with_procfs, uvm_
             channel->channel_info.hwChannelId, channel->channel_info.hwChannelId,
             ce_index);
 
-    status = uvm_rm_locked_call(nvUvmInterfaceCopyEngineAlloc(channel->handle,
-                                                              ce_index,
-                                                              &ce_handle,
-                                                              &channel->channel_info));
-    if (status != NV_OK) {
-        UVM_ERR_PRINT("nvUvmInterfaceCopyEngineAlloc(ce_index=%u) failed: %s, channel %s GPU %s\n",
-                ce_index, nvstatusToString(status), channel->name, gpu->name);
-        goto error;
-    }
-
     channel->gpfifo_entries = uvm_kvmalloc_zero(sizeof(*channel->gpfifo_entries) * channel->num_gpfifo_entries);
     if (channel->gpfifo_entries == NULL) {
         status = NV_ERR_NO_MEMORY;
@@ -575,6 +565,14 @@ static NV_STATUS create_channel(uvm_channel_pool_t *pool, bool with_procfs, uvm_
         goto error;
     }
 
+    if (uvm_push_info_is_tracking_acquires()) {
+        channel->push_acquire_infos = uvm_kvmalloc_zero(sizeof(*channel->push_acquire_infos) * channel->num_gpfifo_entries);
+        if (channel->push_acquire_infos == NULL) {
+            status = NV_ERR_NO_MEMORY;
+            goto error;
+        }
+    }
+
     for (i = 0; i < channel->num_gpfifo_entries; i++)
         list_add_tail(&channel->push_infos[i].available_list_node, &channel->available_push_infos);
 
@@ -613,6 +611,7 @@ static void destroy_channel(uvm_channel_manager_t *manager, uvm_channel_t *chann
     uvm_procfs_destroy_entry(channel->procfs.info);
     uvm_procfs_destroy_entry(channel->procfs.dir);
 
+    uvm_kvfree(channel->push_acquire_infos);
     uvm_kvfree(channel->push_infos);
 
     uvm_kvfree(channel->gpfifo_entries);
@@ -807,13 +806,13 @@ static int compare_ce_for_channel_type(const UvmGpuCopyEngineCaps *ce_caps,
     // Last resort, just order by index
     return ce_index1 - ce_index2;
 }
- 
+
 // Pick default CE for the given channel type, and increment its usage
-// count. This function also sets the i-th bit in usable_ce_mask if the 
-// CE with index i can be used for transfers of the specified type. 
+// count. This function also sets the i-th bit in usable_ce_mask if the
+// CE with index i can be used for transfers of the specified type.
 //
-// Both the usage count array and the CE mask must be initialized prior 
-// to calling this function 
+// Both the usage count array and the CE mask must be initialized prior
+// to calling this function
 static NV_STATUS pick_ce_for_channel_type(uvm_channel_manager_t *manager,
                                           const UvmGpuCopyEngineCaps *ce_caps,
                                           uvm_channel_type_t type,
@@ -869,7 +868,7 @@ static NV_STATUS channel_manager_pick_copy_engines(uvm_channel_manager_t *manage
         return status;
 
    // The order of picking CEs for each type matters as it's affected by the
-   // usage count of each CE and it increases every time a CE is selected. 
+   // usage count of each CE and it increases every time a CE is selected.
    // MEMOPS has the least priority as it only cares about low usage of the
    // CE to improve latency
     for (i = 0; i < ARRAY_SIZE(types); ++i) {
@@ -1133,6 +1132,37 @@ static void uvm_channel_print_info(uvm_channel_t *channel, struct seq_file *s)
     uvm_spin_unlock(&channel->pool->lock);
 }
 
+static void channel_print_push_acquires(uvm_push_acquire_info_t *push_acquire_info, struct seq_file *seq)
+{
+    NvU32 i;
+    NvU32 valid_entries;
+
+    UVM_ASSERT(uvm_push_info_is_tracking_acquires());
+    UVM_ASSERT(push_acquire_info);
+
+    if (push_acquire_info->num_values == 0)
+        return;
+
+    valid_entries = min(push_acquire_info->num_values, (NvU32)UVM_PUSH_ACQUIRE_INFO_MAX_ENTRIES);
+
+    for (i = 0; i < valid_entries; ++i) {
+        UVM_SEQ_OR_DBG_PRINT(seq,
+                             "%s gpu%u:channel%u:%llu",
+                             i == 0? " acquiring values" : "",
+                             push_acquire_info->values[i].gpu_id,
+                             push_acquire_info->values[i].channel_id,
+                             push_acquire_info->values[i].value);
+    }
+
+    if (push_acquire_info->num_values > UVM_PUSH_ACQUIRE_INFO_MAX_ENTRIES) {
+        UVM_SEQ_OR_DBG_PRINT(seq,
+                             " (missing %u entries)",
+                             push_acquire_info->num_values - UVM_PUSH_ACQUIRE_INFO_MAX_ENTRIES);
+    }
+
+    UVM_SEQ_OR_DBG_PRINT(seq, "\n");
+}
+
 // Print all pending pushes and up to finished_pushes_count completed if their
 // GPFIFO entries haven't been reused yet.
 static void channel_print_pushes(uvm_channel_t *channel, NvU32 finished_pushes_count, struct seq_file *seq)
@@ -1149,14 +1179,31 @@ static void channel_print_pushes(uvm_channel_t *channel, NvU32 finished_pushes_c
     for (gpu_get = channel->gpu_get; gpu_get != cpu_put; gpu_get = (gpu_get + 1) % channel->num_gpfifo_entries) {
         uvm_gpfifo_entry_t *entry = &channel->gpfifo_entries[gpu_get];
         uvm_push_info_t *push_info = entry->push_info;
+        uvm_push_acquire_info_t *push_acquire_info = NULL;
 
         if (entry->tracking_semaphore_value + finished_pushes_count <= completed_value)
             continue;
 
-        UVM_SEQ_OR_DBG_PRINT(seq, " %s push '%s' started at %s:%d in %s() releasing value %llu\n",
-                entry->tracking_semaphore_value <= completed_value ? "finished" : "pending",
-                push_info->description, push_info->filename, push_info->line, push_info->function,
-                entry->tracking_semaphore_value);
+        // Obtain the value acquire tracking information from the push_info index
+        if (uvm_push_info_is_tracking_acquires()) {
+            NvU32 push_info_index = push_info - channel->push_infos;
+            UVM_ASSERT(push_info_index < channel->num_gpfifo_entries);
+
+            push_acquire_info = &channel->push_acquire_infos[push_info_index];
+        }
+
+        UVM_SEQ_OR_DBG_PRINT(seq,
+                             " %s push '%s' started at %s:%d in %s() releasing value %llu%s",
+                             entry->tracking_semaphore_value <= completed_value ? "finished" : "pending",
+                             push_info->description,
+                             push_info->filename,
+                             push_info->line,
+                             push_info->function,
+                             entry->tracking_semaphore_value,
+                             !push_acquire_info || push_acquire_info->num_values == 0? "\n" : "");
+
+        if (push_acquire_info)
+            channel_print_push_acquires(push_acquire_info, seq);
     }
     uvm_spin_unlock(&channel->pool->lock);
 }
@@ -1199,7 +1246,7 @@ static int nv_procfs_read_manager_pending_pushes(struct seq_file *s, void *v)
     return 0;
 }
 
-NV_DEFINE_PROCFS_SINGLE_FILE(manager_pending_pushes);
+UVM_DEFINE_SINGLE_PROCFS_FILE(manager_pending_pushes);
 
 static NV_STATUS manager_create_procfs(uvm_channel_manager_t *manager)
 {
@@ -1223,7 +1270,7 @@ static int nv_procfs_read_channel_info(struct seq_file *s, void *v)
     return 0;
 }
 
-NV_DEFINE_PROCFS_SINGLE_FILE(channel_info);
+UVM_DEFINE_SINGLE_PROCFS_FILE(channel_info);
 
 static int nv_procfs_read_channel_pushes(struct seq_file *s, void *v)
 {
@@ -1233,7 +1280,7 @@ static int nv_procfs_read_channel_pushes(struct seq_file *s, void *v)
     return 0;
 }
 
-NV_DEFINE_PROCFS_SINGLE_FILE(channel_pushes);
+UVM_DEFINE_SINGLE_PROCFS_FILE(channel_pushes);
 
 static NV_STATUS channel_create_procfs(uvm_channel_t *channel)
 {
