@@ -1,5 +1,5 @@
 /*******************************************************************************
-    Copyright (c) 2015-2018 NVIDIA Corporation
+    Copyright (c) 2015-2019 NVIDIA Corporation
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"), to
@@ -112,6 +112,10 @@ static NV_STATUS get_gpu_caps(uvm_gpu_t *gpu)
         gpu->sysmem_link = UVM_GPU_LINK_NVLINK_1;
     else if (gpu_caps.sysmemLink == UVM_PEER_LINK_TYPE_NVLINK_2)
         gpu->sysmem_link = UVM_GPU_LINK_NVLINK_2;
+
+
+
+
     else
         UVM_ASSERT(0);
 
@@ -181,9 +185,10 @@ static void gpu_info_print_ce_caps(uvm_gpu_t *gpu, struct seq_file *s)
         if (!ce_caps->supported)
             continue;
 
-        UVM_SEQ_OR_DBG_PRINT(s, " ce %u grce %u shared %u sysmem read %u sysmem write %u sysmem %u nvlink p2p %u "
+        UVM_SEQ_OR_DBG_PRINT(s, " ce %u pce mask 0x%08x grce %u shared %u sysmem read %u sysmem write %u sysmem %u nvlink p2p %u "
                              "p2p %u\n",
                              i,
+                             ce_caps->cePceMask,
                              ce_caps->grce,
                              ce_caps->shared,
                              ce_caps->sysmemRead,
@@ -198,6 +203,8 @@ static void gpu_info_print_common(uvm_gpu_t *gpu, struct seq_file *s)
 {
     NvU64 num_pages_in;
     NvU64 num_pages_out;
+    NvU64 mapped_cpu_pages_size;
+    NvU32 get, put;
 
     UVM_SEQ_OR_DBG_PRINT(s, "GPU %s\n", gpu->name);
     UVM_SEQ_OR_DBG_PRINT(s, "retained_count                         %llu\n",
@@ -279,19 +286,24 @@ static void gpu_info_print_common(uvm_gpu_t *gpu, struct seq_file *s)
                              gpu->access_counter_buffer_info.cached_get);
         UVM_SEQ_OR_DBG_PRINT(s, "access_counters_cached_put             %u\n",
                              gpu->access_counter_buffer_info.cached_put);
-        UVM_SEQ_OR_DBG_PRINT(s, "access_counters_get                    %u\n",
-                             UVM_READ_ONCE(*gpu->access_counter_buffer_info.rm_info.pAccessCntrBufferGet));
-        UVM_SEQ_OR_DBG_PRINT(s, "access_counters_put                    %u\n",
-                             UVM_READ_ONCE(*gpu->access_counter_buffer_info.rm_info.pAccessCntrBufferPut));
+
+        get = UVM_GPU_READ_ONCE(*gpu->access_counter_buffer_info.rm_info.pAccessCntrBufferGet);
+        put = UVM_GPU_READ_ONCE(*gpu->access_counter_buffer_info.rm_info.pAccessCntrBufferPut);
+
+        UVM_SEQ_OR_DBG_PRINT(s, "access_counters_get                    %u\n", get);
+        UVM_SEQ_OR_DBG_PRINT(s, "access_counters_put                    %u\n", put);
     }
 
     num_pages_out = atomic64_read(&gpu->stats.num_pages_out);
     num_pages_in = atomic64_read(&gpu->stats.num_pages_in);
+    mapped_cpu_pages_size = atomic64_read(&gpu->mapped_cpu_pages_size);
 
     UVM_SEQ_OR_DBG_PRINT(s, "migrated_pages_in                      %llu (%llu MB)\n", num_pages_in,
                          (num_pages_in * (NvU64)PAGE_SIZE) / (1024u * 1024u));
     UVM_SEQ_OR_DBG_PRINT(s, "migrated_pages_out                     %llu (%llu MB)\n", num_pages_out,
                          (num_pages_out * (NvU64)PAGE_SIZE) / (1024u * 1024u));
+    UVM_SEQ_OR_DBG_PRINT(s, "mapped_cpu_pages_dma                   %llu (%llu MB)\n", mapped_cpu_pages_size / PAGE_SIZE,
+                         mapped_cpu_pages_size / (1024u * 1024u));
 
     gpu_info_print_ce_caps(gpu, s);
 }
@@ -385,9 +397,9 @@ nv_procfs_read_gpu_access_counters(struct seq_file *s, void *v)
     return 0;
 }
 
-NV_DEFINE_PROCFS_SINGLE_FILE(gpu_info);
-NV_DEFINE_PROCFS_SINGLE_FILE(gpu_fault_stats);
-NV_DEFINE_PROCFS_SINGLE_FILE(gpu_access_counters);
+UVM_DEFINE_SINGLE_PROCFS_FILE(gpu_info);
+UVM_DEFINE_SINGLE_PROCFS_FILE(gpu_fault_stats);
+UVM_DEFINE_SINGLE_PROCFS_FILE(gpu_access_counters);
 
 static NV_STATUS init_procfs_dirs(uvm_gpu_t *gpu)
 {
@@ -682,7 +694,7 @@ static NV_STATUS add_gpu(const NvProcessorUuid *gpu_uuid, uvm_gpu_t **gpu_out)
     }
 
     gpu->arch_hal->init_properties(gpu);
-    uvm_mmu_init_gpu_peer_addresses(gpu);
+    uvm_mmu_init_gpu(gpu);
 
     status = uvm_rm_locked_call(nvUvmInterfaceAddressSpaceCreate(g_uvm_global.rm_session_handle,
                                                                  &gpu->uuid,
@@ -1172,7 +1184,7 @@ uvm_gpu_t *uvm_gpu_get_by_uuid(const NvProcessorUuid *gpu_uuid)
     return uvm_gpu_get_by_uuid_locked(gpu_uuid);
 }
 
-NV_STATUS uvm_gpu_retain_by_uuid_locked(const NvProcessorUuid *gpu_uuid, uvm_gpu_t **gpu_out)
+static NV_STATUS gpu_retain_by_uuid_locked(const NvProcessorUuid *gpu_uuid, uvm_gpu_t **gpu_out)
 {
     NV_STATUS status = NV_OK;
     uvm_gpu_t *gpu;
@@ -1195,7 +1207,7 @@ NV_STATUS uvm_gpu_retain_by_uuid(const NvProcessorUuid *gpu_uuid, uvm_gpu_t **gp
 {
     NV_STATUS status;
     uvm_mutex_lock(&g_uvm_global.global_lock);
-    status = uvm_gpu_retain_by_uuid_locked(gpu_uuid, gpu_out);
+    status = gpu_retain_by_uuid_locked(gpu_uuid, gpu_out);
     uvm_mutex_unlock(&g_uvm_global.global_lock);
     return status;
 }
@@ -1239,6 +1251,9 @@ void uvm_gpu_release_mask_locked(const uvm_processor_mask_t *mask)
 
 void uvm_gpu_release_mask(const uvm_processor_mask_t *mask)
 {
+    if (uvm_processor_mask_empty(mask))
+        return;
+
     uvm_mutex_lock(&g_uvm_global.global_lock);
     uvm_gpu_release_mask_locked(mask);
     uvm_mutex_unlock(&g_uvm_global.global_lock);
@@ -1434,6 +1449,10 @@ static NV_STATUS init_peer_access(uvm_gpu_t *gpu_1,
         peer_caps->link_type = UVM_GPU_LINK_NVLINK_1;
     else if (p2p_caps_params->p2pLink == UVM_PEER_LINK_TYPE_NVLINK_2)
         peer_caps->link_type = UVM_GPU_LINK_NVLINK_2;
+
+
+
+
     else
         return NV_ERR_NOT_SUPPORTED;
 
@@ -1869,7 +1888,7 @@ static NV_STATUS gpu_add_user_channel_subctx_info(uvm_gpu_t *gpu, uvm_user_chann
     // below
     if (new_channel_subctx_info) {
         new_channel_subctx_info->subctxs =
-            uvm_kvmalloc_zero(sizeof(*new_channel_subctx_info->subctxs) * gpu->max_subcontexts);
+            uvm_kvmalloc_zero(sizeof(*new_channel_subctx_info->subctxs) * user_channel->tsg.max_subctx_count);
     }
 
     // Pre-load the tree to allocate memory outside of the table lock. This
@@ -1884,7 +1903,7 @@ static NV_STATUS gpu_add_user_channel_subctx_info(uvm_gpu_t *gpu, uvm_user_chann
 
     // Check if the subcontext information for the channel already exists
     channel_subctx_info =
-        (uvm_user_channel_subctx_info_t *)radix_tree_lookup(&gpu->tsg_table, user_channel->tsg_id);
+        (uvm_user_channel_subctx_info_t *)radix_tree_lookup(&gpu->tsg_table, user_channel->tsg.id);
 
     if (!channel_subctx_info) {
         // We could not allocate the descriptor before taking the lock. Exiting
@@ -1894,10 +1913,16 @@ static NV_STATUS gpu_add_user_channel_subctx_info(uvm_gpu_t *gpu, uvm_user_chann
         }
 
         // Insert the new subcontext information descriptor
-        ret = radix_tree_insert(&gpu->tsg_table, user_channel->tsg_id, new_channel_subctx_info);
+        ret = radix_tree_insert(&gpu->tsg_table, user_channel->tsg.id, new_channel_subctx_info);
         UVM_ASSERT(ret == 0);
 
         channel_subctx_info = new_channel_subctx_info;
+
+
+
+
+
+
     }
 
     user_channel->subctx_info = channel_subctx_info;
@@ -1911,7 +1936,7 @@ static NV_STATUS gpu_add_user_channel_subctx_info(uvm_gpu_t *gpu, uvm_user_chann
                        instance_ptr.address,
                        uvm_aperture_string(instance_ptr.aperture),
                        user_channel->subctx_id,
-                       user_channel->tsg_id,
+                       user_channel->tsg.id,
                        (NvU64)va_space,
                        (NvU64)channel_subctx_info->subctxs[user_channel->subctx_id].va_space);
         UVM_ASSERT_MSG(channel_subctx_info->subctxs[user_channel->subctx_id].va_space != NULL,
@@ -1920,14 +1945,14 @@ static NV_STATUS gpu_add_user_channel_subctx_info(uvm_gpu_t *gpu, uvm_user_chann
                        instance_ptr.address,
                        uvm_aperture_string(instance_ptr.aperture),
                        user_channel->subctx_id,
-                       user_channel->tsg_id);
+                       user_channel->tsg.id);
         UVM_ASSERT_MSG(channel_subctx_info->total_refcount > 0,
                        "CH %u instance_ptr {0x%llx:%s} SubCTX %u in TSG %u: TSG refcount is 0\n",
                        user_channel->hw_channel_id,
                        instance_ptr.address,
                        uvm_aperture_string(instance_ptr.aperture),
                        user_channel->subctx_id,
-                       user_channel->tsg_id);
+                       user_channel->tsg.id);
     }
     else {
         UVM_ASSERT_MSG(channel_subctx_info->subctxs[user_channel->subctx_id].va_space == NULL,
@@ -1936,7 +1961,7 @@ static NV_STATUS gpu_add_user_channel_subctx_info(uvm_gpu_t *gpu, uvm_user_chann
                        instance_ptr.address,
                        uvm_aperture_string(instance_ptr.aperture),
                        user_channel->subctx_id,
-                       user_channel->tsg_id,
+                       user_channel->tsg.id,
                        (NvU64)channel_subctx_info->subctxs[user_channel->subctx_id].va_space);
 
         channel_subctx_info->subctxs[user_channel->subctx_id].va_space = va_space;
@@ -1977,7 +2002,7 @@ static void gpu_remove_user_channel_subctx_info_locked(uvm_gpu_t *gpu, uvm_user_
     // Channel subcontext info descriptor may not have been registered in
     // tsg_table since this function is called in some teardown paths during
     // channel creation
-    channel_subctx_info = (uvm_user_channel_subctx_info_t *)radix_tree_lookup(&gpu->tsg_table, user_channel->tsg_id);
+    channel_subctx_info = (uvm_user_channel_subctx_info_t *)radix_tree_lookup(&gpu->tsg_table, user_channel->tsg.id);
     UVM_ASSERT(channel_subctx_info == user_channel->subctx_info);
 
     UVM_ASSERT_MSG(channel_subctx_info->subctxs[user_channel->subctx_id].refcount > 0,
@@ -1986,7 +2011,7 @@ static void gpu_remove_user_channel_subctx_info_locked(uvm_gpu_t *gpu, uvm_user_
                    instance_ptr.address,
                    uvm_aperture_string(instance_ptr.aperture),
                    user_channel->subctx_id,
-                   user_channel->tsg_id);
+                   user_channel->tsg.id);
 
     UVM_ASSERT_MSG(channel_subctx_info->subctxs[user_channel->subctx_id].va_space == va_space,
                    "CH %u instance_ptr {0x%llx:%s} SubCTX %u in TSG %u: expected VA space 0x%llx but got 0x%llx instead\n",
@@ -1994,7 +2019,7 @@ static void gpu_remove_user_channel_subctx_info_locked(uvm_gpu_t *gpu, uvm_user_
                    instance_ptr.address,
                    uvm_aperture_string(instance_ptr.aperture),
                    user_channel->subctx_id,
-                   user_channel->tsg_id,
+                   user_channel->tsg.id,
                    (NvU64)va_space,
                    (NvU64)channel_subctx_info->subctxs[user_channel->subctx_id].va_space);
 
@@ -2004,21 +2029,21 @@ static void gpu_remove_user_channel_subctx_info_locked(uvm_gpu_t *gpu, uvm_user_
                    instance_ptr.address,
                    uvm_aperture_string(instance_ptr.aperture),
                    user_channel->subctx_id,
-                   user_channel->tsg_id);
+                   user_channel->tsg.id);
 
     // Decrement VA space refcount. If it gets to zero, unregister the pointer
     if (--channel_subctx_info->subctxs[user_channel->subctx_id].refcount == 0)
         channel_subctx_info->subctxs[user_channel->subctx_id].va_space = NULL;
 
     if (--channel_subctx_info->total_refcount == 0) {
-        channel_subctx_info = (uvm_user_channel_subctx_info_t *)radix_tree_delete(&gpu->tsg_table, user_channel->tsg_id);
+        channel_subctx_info = (uvm_user_channel_subctx_info_t *)radix_tree_delete(&gpu->tsg_table, user_channel->tsg.id);
         UVM_ASSERT_MSG(channel_subctx_info == user_channel->subctx_info,
                        "CH %u instance_ptr {0x%llx:%s} SubCTX %u in TSG %u: subctx info found: 0x%llx, but expected: 0x%llx\n",
                        user_channel->hw_channel_id,
                        instance_ptr.address,
                        uvm_aperture_string(instance_ptr.aperture),
                        user_channel->subctx_id,
-                       user_channel->tsg_id,
+                       user_channel->tsg.id,
                        (NvU64)channel_subctx_info,
                        (NvU64)user_channel->subctx_info);
     }
@@ -2071,7 +2096,7 @@ static NV_STATUS gpu_add_user_channel_instance_ptr(uvm_gpu_t *gpu, uvm_user_chan
                    instance_ptr.address,
                    uvm_aperture_string(instance_ptr.aperture),
                    user_channel->subctx_id,
-                   user_channel->tsg_id,
+                   user_channel->tsg.id,
                    ret);
 
     return NV_OK;
@@ -2136,17 +2161,17 @@ static uvm_va_space_t *user_channel_and_subctx_to_va_space(uvm_user_channel_t *u
 
     channel_subctx_info = user_channel->subctx_info;
 
-    UVM_ASSERT_MSG(subctx_id < user_channel->gpu->max_subcontexts,
+    UVM_ASSERT_MSG(subctx_id < user_channel->tsg.max_subctx_count,
                    "instance_ptr {0x%llx:%s} in TSG %u. Invalid SubCTX %u\n",
                    user_channel->instance_ptr.address,
                    uvm_aperture_string(user_channel->instance_ptr.aperture),
-                   user_channel->tsg_id,
+                   user_channel->tsg.id,
                    subctx_id);
     UVM_ASSERT_MSG(channel_subctx_info->total_refcount > 0,
                    "instance_ptr {0x%llx:%s} in TSG %u: TSG refcount is 0\n",
                    user_channel->instance_ptr.address,
                    uvm_aperture_string(user_channel->instance_ptr.aperture),
-                   user_channel->tsg_id);
+                   user_channel->tsg.id);
 
     // A subcontext's refcount can be zero if that subcontext is torn down
     // uncleanly and work from that subcontext continues running with work from
@@ -2159,7 +2184,7 @@ static uvm_va_space_t *user_channel_and_subctx_to_va_space(uvm_user_channel_t *u
                        "instance_ptr {0x%llx:%s} in TSG %u: no VA space for SubCTX %u\n",
                        user_channel->instance_ptr.address,
                        uvm_aperture_string(user_channel->instance_ptr.aperture),
-                       user_channel->tsg_id,
+                       user_channel->tsg.id,
                        subctx_id);
     }
 
@@ -2196,7 +2221,16 @@ NV_STATUS uvm_gpu_fault_entry_to_va_space(uvm_gpu_t *gpu, uvm_fault_buffer_entry
         *out_va_space = user_channel->gpu_va_space->va_space;
     }
     else {
-        *out_va_space = user_channel_and_subctx_to_va_space(user_channel, fault->fault_source.ve_id);
+        NvU32 ve_id = fault->fault_source.ve_id;
+
+
+
+
+
+
+
+
+        *out_va_space = user_channel_and_subctx_to_va_space(user_channel, ve_id);
 
         // Instance pointer is valid but the fault targets a non-existent
         // subcontext.
@@ -2337,7 +2371,7 @@ NV_STATUS uvm_gpu_map_cpu_pages(uvm_gpu_t *gpu, struct page *page, size_t size,
 
     UVM_ASSERT(PAGE_ALIGNED(size));
 
-    if (NV_PCI_DMA_MAPPING_ERROR(gpu->pci_dev, dma_addr))
+    if (pci_dma_mapping_error(gpu->pci_dev, dma_addr))
         return NV_ERR_OPERATING_SYSTEM;
 
     if (dma_addr < gpu->dma_addressable_start || dma_addr + size - 1 > gpu->dma_addressable_limit) {
@@ -2453,12 +2487,15 @@ NV_STATUS uvm_api_pageable_mem_access_on_gpu(UVM_PAGEABLE_MEM_ACCESS_ON_GPU_PARA
 
 NV_STATUS uvm8_test_set_prefetch_filtering(UVM_TEST_SET_PREFETCH_FILTERING_PARAMS *params, struct file *filp)
 {
+    uvm_va_space_t *va_space = uvm_va_space_get(filp);
     uvm_gpu_t *gpu = NULL;
     NV_STATUS status = NV_OK;
 
     uvm_mutex_lock(&g_uvm_global.global_lock);
 
-    gpu = uvm_gpu_get_by_uuid(&params->gpu_uuid);
+    uvm_va_space_down_read(va_space);
+
+    gpu = uvm_va_space_get_gpu_by_uuid(va_space, &params->gpu_uuid);
     if (!gpu) {
         status = NV_ERR_INVALID_DEVICE;
         goto done;
@@ -2477,26 +2514,28 @@ NV_STATUS uvm8_test_set_prefetch_filtering(UVM_TEST_SET_PREFETCH_FILTERING_PARAM
     }
 
 done:
+    uvm_va_space_up_read(va_space);
+
     uvm_mutex_unlock(&g_uvm_global.global_lock);
     return status;
 }
 
 NV_STATUS uvm8_test_get_gpu_time(UVM_TEST_GET_GPU_TIME_PARAMS *params, struct file *filp)
 {
+    uvm_va_space_t *va_space = uvm_va_space_get(filp);
     uvm_gpu_t *gpu = NULL;
     NV_STATUS status = NV_OK;
 
-    uvm_mutex_lock(&g_uvm_global.global_lock);
+    uvm_va_space_down_read(va_space);
 
-    gpu = uvm_gpu_get_by_uuid(&params->gpu_uuid);
-    if (!gpu) {
+    gpu = uvm_va_space_get_gpu_by_uuid(va_space, &params->gpu_uuid);
+
+    if (gpu)
+        params->timestamp_ns = gpu->host_hal->get_time(gpu);
+    else
         status = NV_ERR_INVALID_DEVICE;
-        goto done;
-    }
 
-    params->timestamp_ns = gpu->host_hal->get_time(gpu);
+    uvm_va_space_up_read(va_space);
 
-done:
-    uvm_mutex_unlock(&g_uvm_global.global_lock);
     return status;
 }
